
/*
**  dLNN       :
**
**  purpose   : dNN produces a gradient vector of a linear/neural network lnn
**              (a 3 layer neural network);  given input data x,
**              produces an output vector ( matrix?) y.
**
**  syntax    : dy = dlnn(X,A,b,C) ;
**
**  input     :
**              x :   T x k   matrix of input values
**              A :   k x nH  matrix of connection weights between
**                            input layer and hidden layer.
**              b :   k x 1   linear coefficients
**              C :  nH x 1?  vector of connection weights between input
**                            layer and output layer
**
**  output    : dy     : T x q  : gradient of output of neural netwerk,
**                                evaluated at each input x.
**                                One row of dy corresponds to the
**                                gradient with respect to A,b,C of the
**                                network function nn;
**                                q = dim(A)+dim(b)+dim(C);
**                                a reshape of dy[i] produces dy/dA,
**                                dy/db,dy/dC evaluated at x[i];
**  remarks   :
**               This implementation uses g(x)= 1/(1+exp(-x))
**               as the activation function;
**               change the dimension of C to nH x nO to produce
**               more output values.
**
**
**  date      :  16-11-93
**
**  programmer: G. Draisma
**
*/

proc dlnn(x,A,b,C) ;

local
    g, dg, nH, k, i, T, dA, dB, dC ;

T     =  rows(X) ;
k     =  cols(X) ;
nH    =  cols(A) ;
dA    =  zeros(T, k*nH) ;
g     =  1./(1+exp(-(x*A))) ;
dC    =  g ;
dg    =  (g.*(1-g))  .* C' ;
i = 0 ;
do while i < k ;
   i = i + 1 ;
   dA[.,(nH*(i-1)+1):nH*i] = dg.*X[.,i] ;
endo ;
dB = x ;
retp((dA~dB~dC)) ;
endp ;





/*
**  dNN       :
**
**  syntax    : {dy}   = dcg(x,A,C) ;
**
**  purpose   : dNN produces a gradient vector of a neural network nn
**              (a 3 layer neural network);  given input data x,
**              produces an output vector ( matrix?) y.
**
**  syntax    : dy = dnn(X,A,b,C) ;
**
**  input     :
**              x :   T x k   matrix of input values
**              A :   k x nH matrix of connection weights between
**                            input layer and hidden layer.
**              (b :   1 x nH vector ("hidden input cell, or thresholds"))
**              included in x dummies
**              C :  nH x 1?  vector of connection weights between input
**                            layer and output layer
**
**  output    : dy     : T x q  : gradient of output of neural netwerk,
**                                evaluated at each input x.
**                                One row of dy corresponds to the
**                                gradient with respect to A,b,C of the
**                                network function nn;
**                                q = dim(A)+dim(b)+dim(C);
**                                a reshape of dy[i] produces dy/dA,
**                                dy/db,dy/dC evaluated at x[i];
**  remarks   :
**               This implementation uses g(x)= 1/(1+exp(-x))
**               as the activation function;
**               change the dimension of C to nH x nO to produce
**               more output values.
**
**
**  date      :  16-11-93
**
**  programmer: G. Draisma
**
*/

proc dnn(x,A,C) ;

local
    g, nH, k, i, T, dA, dB, dC ;

T     =  rows(X) ;
k     =  cols(X) ;
nH    =  cols(A) ;
dA    =  zeros(T, k*nH) ;
g     =  1./(1+exp(-(x*A))) ;
dC    =  g ;
dB    =  (g.*(1-g))  .* C' ;
i = 0 ;
do while i < k ;
   i = i + 1 ;
   dA[.,(nH*(i-1)+1):nH*i] = dB.*X[.,i] ;
endo ;
retp((dA~dC)) ;
endp ;




/*
**  lnn.e
**
**  purpose   : demonstration of a simple 3 layer neural network:
**              y : data
**              x : lagged data + season dummy
**
**  globals   : this program uses the following subroutines:
**              0. nelder (simplex algorithm)
**              1. Optmum (library optmum)
**              2. lnn(X,A,b,C)  : output of the neural network
**              3. dlnn(X,A,b,C) : gradient: dlnn/dA,..., used for optimizing
**                 the network
**              The data is represented
**              by the globals:
**                  _y      : (n x 1) vector dependent
**                  _x      : (n x k) matrix independent vars.
**
**  remarks   : The network is consists of:
**                  a linear part
**                  an input layer with   k cells
**                  a  hidden layer with nH cells
**                  an output layer with  m=1 cells
**
**
**              It's current learning state is represented by:
**                  A : (k x nH) connections weights input -> hidden
**                  b :  k x 1   coefficients of the linear part
**
**                  C : (nH x m) connection weights hidden -> output
**
**              Learning consists of minimizing the sum of squares:
**                  sum((_y - lnn(_x,A,b,C))^2)+1 over A,b,C
**
**              In this example the network is used to locate
**              trend-breaks by supplementing a linear network
**              with a nonlinear neural network function.
**
**  date      : 30-07-94
**
**  programmer: G. Draisma
**
*/

/*
** these are weight decay coefs
*/
declare _rA != 1E-8 ;
declare _rB != 1E-4 ;
declare _rC != 1E-4 ;

library optmum, pgraph ;
format /rz 8, 4 ;

plotnr = 0 ;


try_nH = { 1,2,3 } ;
idat = 1 ;

let datfil =
    UR URA;
let cname =
    UR URA;


let show_nH =
    1   1  ;

LOOPDAT: ;  /* start of loop for each data file */
/*
**  read in data : define globals
*/
itry = 1 ;
fname  = "" $+ datfil[idat] $+ ".asc" ;
outfil = "" $+ datfil[idat] $+ "x.tmp" ;
output file = ^outfil reset ;
output on ;
load x[] = ^fname ;
T  = rows(x) ;
s1  = seqa(0,1,T)%4;
s   = dummy(s1,0|1|2) ;
jr = seqa(62+1/8,1/4,T) ;           /* time series */

/* scale data to [0,1] : facilitates fitting */
x  = (x-minc(x))./(maxc(x)-minc(x)) ;
_y = x ;

/*
** define the variables (_x) to use for explanation:
**    4 lags (seasons)
**    4 season dummies (1 2 3 4)
**    1 trend variable
*/
v  = { 1,2,3,4 } ;
x  = shiftr(x',v,0)' ;
_y = _y[maxc(v)+1:T]   ;
_x =  x[maxc(v)+1:T,.] ;
_x = _x~s[maxc(v)+1:T,.]~seqa(0,1/(rows(_x)-1),rows(_x)) ;
jr = jr[maxc(v)+1:T] ;
s  = s[maxc(v)+1:T,.] ;

LOOPnH: /* start  of for each number of hidden cells */

/*
** dimensions of network, and indexes theta <-> A,b,c
*/
savfil = "" $+ datfil[idat] $+ ftos(try_nH[itry],"L*.*lfX",1,0) ;
_nH =  try_nH[itry] ;
_nO =  1 ;
_nI = cols(_x) ;
_iA = seqa(1,1,_nI*_nH) ;
_ib = seqa(_nI*_nH+1,1,_nI) ;
_ic = seqa(_nI*(_nH+1)+1,1,_nH) ;
fth1 = 1000 ;
th1  = 0 ;

iover = 0 ;
/* start of loop for retries */
OVERNIEW:
iover = iover + 1 ;

/*
** Start values:
*/
A  = rndu(_nI,_nH)-0.5 ;
b  = _y/_x  ;
c  = rndu(_nH,1)-0.5   ;
th =  vecr(A)|b|c  ;

/*
** NELDER:
** expand starting point to a complete simplex
** add a (signed) I-matrix to ensure that the start simplex
** includes the origin in some way.
*/
th = th~(2*((th.<0)-0.5).*(eye(rows(th)))+th) ;

/* remove comments to restart optimization at last point */
@ loadm th = ^savfil ; @

/* number of iterations for optmum */
iters = (100|200|200|200)*_nH ;
nrounds = rows(iters) ;


/* OPT-close / SHOW-open -> * /

/*
** use the simplex method for finding a starting point
** number of iterations dependent on problem size
*/
_maxit = 100 + 10*rows(th) ;
_ftol = 1E-6 ;
_show = 50 ;
format /rz 9, 4 ;
$datfil[idat] ;
{th,fth} = nelder(&err,th) ;

/* check the derivative procedure
** against a numeric test
*/
"test derivative:"    ;
_fcmptol = 1E-3*maxc(derr(th)') ;
feq(gradp(&err,th),derr(th)) ;


_opgdprc = &dErr ;
_opgtol  = 1E-6 ;
__output = 0 ;
ri = 0 ;
do while ri < nrounds ;
    ri = ri + 1 ;
    _opmiter = iters[ri] ;
    "ri" ri ;
    datestr(date) ; timestr(time) ;
    {th, fth, dth, retcode} = optmum(&Err,th) ;
    reshape(th,rows(th)/_nH,_nH) ;
    "ssr " ; fth-1 ;
endo ;
datestr(date) ; timestr(time) ;
ssr  = err(vecr(th));
ssr  = ssr-1-_rA*sumc(th[_iA,.]^2)-_rB*sumc(th[_iB,.]^2)-_rC*sumc(th[_iC,.]^2) ;
"ssr           "  ssr ;
CS             =  (rows(_x)*ln(ssr/rows(_x)) + rows(th)*ln(rows(_x)));
"CS            "  cs ;
if fth1 > fth  ;
  th1 = th ;
  fth1 = fth ;
endif ;

if iover < 3 ; /* put here a number of retries */
   goto OVERNIEW ;
endif ;

/* save the best results */
save ^savfil = th1 ;

itry = itry + 1 ;
if itry <= rows(try_nH) ;
   goto LOOPnH ;
endif ;
if idat < rows(datfil) ;
   idat = idat + 1 ;
   goto LOOPDAT ;
endif ;

end ;

/ * <- OPT-close /SHOW-open  -> */

/*
** Below is for presenting some results:
** plot's and parameters.
** the results from the optimization are read in
** and then parameters printed and
** network performance plotted
*/

/*
** this is to select only the plots specified
** in show_nH
*/
@
if _nH /= show_nH[idat] ;
   goto NOPLOT ;
endif;
@

"read in " savfil ;
loadm th1 = ^savfil ;
A   = reshape(th1[_iA],_nI,_nH) ;
b   = th1[_ib] ;
C   = th1[_ic] ;

/* Output parameters naar log-file */
output ON ;
" " ;
"Linear Neural Network" ;
"read in " savfil ;
"Hidden layers " _nH ;
"Iterations    " sumc(iters) ;
"Lags          "  v' ;
"Network parameters" ;
"b (lin coefs) " ; b' ;
"A " ; A' ;
"C'" ; C' ;
ssr  = err(vecr(th1)) ;
ssr  = ssr-1-_rA*sumc(th1[_iA,.]^2)-_rB*sumc(th1[_iB,.]^2)
          -_rC*sumc(th1[_iC,.]^2) ;
"ssr           "  ssr ;
"CS            "  (rows(_x)*ln(ssr/rows(_x)) + rows(th1)*ln(rows(_x)));

@@ wait ;@@

output OFF ;
@ continue ; @
@ goto NOPLOT ; @
@ goto HIDDEN ; @


/*
**  network output
*/
NETWORK:
fonts("complex") ;
graphset ;
plotnr = plotnr+1 ;
_ptek = ftos(plotnr,"fig%*.*lf.tkf",1,0) ;
@ goto HIDDEN ; @
_plwidth = { 1,2 } ;
_pstype = {2,4}  ;
@ _pnum = 0 ; @
_psymsiz = {2,2} ;
_pltype =  { 2,6 } ;
_plctrl = { 1 } ;
_pcolor = { 10, 14 } ;
_pcross = 1 ;
_pframe = 0 ;
_ptitlht = 0.4 ;
begwind ;
window(3,2,0) ;
tit = "" $+ datfil[idat] $+ " with Lnn(" $+ ftos(_nI, "*.*lf,",1,0)
             $+ ftos(_nH, "*.*lf,",1,0)
             $+ ftos(_nO, "*.*lf) prediction ",1,0) ;
tit1 = tit ;
title(tit1) ;
xlabel("T") ;
xtics(60,95,5,5) ;
ytics(0,1,0.5,5) ;
setwind(1) ;
xy(jr,_y~lnn(_x,A,b,C)) ;
title("") ;
is = 0 ;
do while is < 4 ;
    is = is + 1 ;
    vs = indexcat(s[.,is],1) ;
    setwind(2 + is) ;
    xy(jr[vs],_y[vs]~lnn(_x[vs,.],A,b,C)) ;
endo ;
endwind ;

/*
**  input layers:
**  to investigate contribution of each input variable,
**  the effect of leaving that variable is shown:
**  inputs that do not contribute anything, can be
**  left out without effecting the result.
*/

INPUT:
graphset ;
plotnr = plotnr+1 ;
_ptek = ftos(plotnr,"fig%*.*lf.tkf",1,0) ;
_plwidth = { 1,2 } ;
_pstype = {2,4}  ;
@ _pnum = 0 ; @
_psymsiz = {2,2} ;
_pltype =  { 2,6 } ;
_plctrl = { 1 } ;
_pcolor = { 10, 14 } ;
_pnum = 0 ;
_ptitlht = 0.4 ;
_paxht = 0.8 ;
_pcross = 1 ;
_pframe = 0 ;
vv = seqa(1,1,_nI) ;
begwind ;
   window(_nI,1,0) ;
   i = 0 ;
   xtics(60,95,1,0) ;
   do while i < _nI ;
       i = i + 1 ;
       setwind(i) ;
       tit = ftos(i,"I]*.*lf[", 1,0) ;
       ylabel(tit) ;
       ix = delif(vv,vv.==i) ;
       xy(jr,_y~lnn(_x[.,ix],A[ix,.],b[ix],C)) ;
   endo ;
endwind ;

/*
  hidden layers
*/

HIDDEN:
graphset ;
plotnr = plotnr+1 ;
_ptek = ftos(plotnr,"fig%*.*lf.tkf",1,0) ;
_plwidth = { 1,2 } ;
_pstype = {2,4}  ;
@ _pnum = 0 ; @
_psymsiz = {2,2} ;
_pltype =  { 2,6 } ;
_plctrl = { 1 } ;
_pcolor = { 10, 14 } ;
_paxht = 0.8 ;
_pcross = 1 ;
i   = 0 ;
b0  = b.*0 ;
vv  = eye(_nH) ;
@ vv   = vv ~ones(_nH,1) ; @
@ vv   = vv[.,1|2|3|5] ; vv[4,3] = 1 ; @
@ vv = eye(_nH-1)|zeros(1,_nH-1); vv[6,2] = 1 ; @
begwind ;
    makewind(9,0.5,0,6.2,0) ;
    _paxes = 0 ;
    _pnum  = 0 ;
    _ptitlht = 2.5 ;
    title("Output hidden layer cells in each quarter: "
           $+ cname[idat]) ;
    draw ;
    _ptitlht = 0.4 ;
    window(try_nH[rows(try_nH)]+1,4,0) ;
    is = 0 ;
    title("") ;
    _pmsgctl = (7.5~0.3~0.6~0~2~14~0) ;
    do while is < 4 ;
        is = is + 1 ;
        _pmsgstr = ftos(is,"Q]*.*lf[",1,0) ;
        setwind(is+1) ;
        draw ;
    endo ;
    _paxes = 1 ;
    _pnum  = {1,0} ;
    _pnumht = 0.3 ;
    _ptitle = "" ;
    _pframe = 0 ;
    @ window(cols(vv)+1,1,0) ; @
    xtics(60,90,5,5) ;
    ytics(-1,1,1,0) ;
    ticmin = minc(_y|c|-0.5) ;
    ticmax = maxc(_y|c|0.5) ;
    ytics(ticmin,ticmax,ticmax-ticmin,0) ;
    ylabel("b") ;
    ylabel("") ;
    do while i < cols(vv) ;
        i = i + 1 ;
        ix = indexcat(vv[.,i],1) ;
        j = 1 ;
        tit = "H]" $+ftos(ix[j],"*.*lf",1,0) ;
        do while j < rows(ix) ;
           j = j + 1 ;
           tit = tit $+ftos(ix[j],",*.*lf",1,0) ;
        endo ;
        tit = tit $+ "[" ;
        _pmsgstr = tit ;
        _pmsgctl = (0.6~0.3~0.6~0~2~14~0) ;
        ylabel("")  ;
        is = 0 ;
        do while is < 4 ;
            is = is + 1 ;
            vs = indexcat(s[.,is],1) ;
            setwind((i)*4 +is+1) ;
            xy(jr[vs],_y[vs]~lnn(_x[vs,.],A[.,ix],b0,C[ix])) ;
            _pmsgstr = "" ;
        endo ;
    endo ;
endwind ;

NOPLOT:

itry = itry + 1 ;
if itry <= rows(try_nH) ;
   goto LOOPnH ;
endif ;
if idat < rows(datfil) ;
   idat = idat + 1 ;
   goto LOOPDAT ;
endif ;



proc Err(th) ;
local A,b,C,i,f ;
   i = 0 ;
   f = reshape(0,cols(th),1) ;
   do while i < cols(th) ;
      i   = i + 1 ;
      A   = reshape(th[_iA,i],_nI,_nH) ;
      b   = th[_ib,i] ;
      C   = th[_ic,i] ;
      f[i] = (sumc((lnn(_x,A,b,C)-_y)^2)) ;
   endo ;
   retp(f+
    1+_rA*sumc(th[_iA,.]^2) +_rB*sumc(th[_iB,.]^2)+_rC*sumc(th[_iC,.]^2));
endp ;

proc dErr(th) ;
   local A,b,C,i,f ;
      i   = 1 ;
      A   = reshape(th[_iA,i],_nI,_nH) ;
      b   = th[_ib,i] ;
      C   = th[_ic,i] ;
      f   = (sumc(2*(lnn(_x,A,b,C)-_y).*dlnn(_x,A,b,C))') ;
      f   = f + 2 *(_rA*th[_iA,i]|_rB*th[_iB,i]|_rC*th[_iC,i])' ;
     retp(f) ;
endp ;


/*

declare _rA != 1E-6 ;
declare _rC != 1E-2 ;

proc Err(th) ;
local A,C,i,f ;
   i = 0 ;
   f = reshape(0,cols(th),1) ;
   do while i < cols(th) ;
      i   = i + 1 ;
      A   = reshape(th[_iA,i],_nI,_nH) ;
      C   = th[_ic,i] ;
      f[i] = (sumc((nn(_x,A,C)-_y)^2)+1) ;
   endo ;
   retp(f+_rA*sumc(th[_iA,.]^2)+_rC*sumc(th[_iC,.]^2)) ;
endp ;

proc dErr(th) ;
   local A,C,i,f ;
      i   = 1 ;
      A   = reshape(th[_iA,i],_nI,_nH) ;
      C   = th[_ic,i] ;
      f   = (sumc(2*(nn(_x,A,C)-_y).*dnn(_x,A,C))') ;
      f   = f + 2 *(_rA*th[_iA,i]|_rC*th[_iC,i])' ;
     retp(f) ;
endp ;

*/

/*
**  LNN       : neural network
**
**  purpose   : LNN is a linear 3 layer neural network which given input data x,
**              produces an output vector ( matrix?) y.
**
**
**  syntax    : y = nn(X,A,b,c) ;
**
**  input     :
**              x :   T x k   matrix of input values
**              A :   k x nH  matrix of connection weights between
**                            input layer and hidden layer.
**              c :  nH x 1?  vector of connection weights between input
**                            layer and output layer
**              b :   k x 1   linear coefficients
**
**  output    : y :   T x 1?  : output of neural netwerk, with
**                                y = X*b + G(X*A) * C, with G being
**                                the activation function
**
**
**  remarks   :
**               This implementation uses g(x)= 1/(1+exp(-x))
**               as the activation function for the nonlinear part
**               change the dimension of C to nH x nO to produce
**               more output values.
**
**
**  date      :  30-07-94
**
**  programmer: G. Draisma
**
*/

proc lnn(X,A,b,C) ;
    retp(X*b+(1./(1+exp(-(X*A))))*C) ;
endp ;

/*
**   NELDER
**
**   purpose: multivariate minimization routine
**
**   format:  { xmin, fret } = nelder( &f, x) ;
**
**   input:   &f : function or proc returning a scaler
**            f should be defined as
**            { y } = f(theta) ;
**            with theta : (k x 1) vector or
**                         (k x n) matrix
**                  y    :  scalar or (n x 1) vector of
**                          corresponding function values
**
**             x : (k x 1) start value (start simplex based on x)
**                 (k x n+1) start simplex
**
**
**   output:  xmin : (k x 1) localized minimum
**            fret : scalar = f(xmin)
**
**   globals: _ftol : tolerance (relative difference in function values)
**            _maxit: maximum number of iterations
**            _show:  -1: dont show, n: show results every n iterations
**
**   remarks:
**             See W.H. Press (1986) et all,
**             Numerical Recipes: Ch.10 : the Downhill Simplex Method
**             Cambridge University Press
**
**             User can specify the start simplex!
**             implementation:    G. Draisma  25-09-92
**                                Econometric Institute
**                                Erasmus University Rotterdam
**
*/

declare  _ftol  ?= 1.0E-6 ;
declare  _maxit ?= 500 ;
declare  _show  ?= 25 ;

proc (2) = nelder( &f, x ) ;
   local ky, alp, bet, gam,
      del, fp, dim, ndx,
      ilo, ihi, inhi,
      iter,
      p, pbar,
      xp, fxp,
      xxp, fxxp,
      f:proc ;

   p = x ;
   if cols(p) /= 1 and cols(p) /= rows(p)+1 ;
      errorlog "nelder: x -NO- column vector /or simplex" ;
      retp(p,0) ;
   endif ;

   alp  = 1.5 ;  /* mirror factor */
   bet  = 0.4 ;  /* shrink factor */
   gam  = 4.0 ;  /* extend-on-succes factor */
   iter  = 0 ;
   dim  = rows(p) ;
   if  cols(p) == 1 ;
       p    = (p)~(p + (0.1*eye(dim)).*(p+sqrt(_ftol))) ;
   endif ;
   del  = 1 ;
   do while (del > _ftol) and (iter <= _maxit);
        iter  = iter + 1 ;
        fp   = f(p);
        ndx  = sortind(fp) ;
        ihi  = ndx[dim+1] ;
        inhi = ndx[dim] ;
        ilo  = ndx[1] ;
        del  = abs(fp[ihi] - fp[ilo])./(abs(fp[ilo]) + _ftol);
        pbar = meanc(p[.,ndx[1:dim]]') ;
        if (iter % _show == 1) ;
           ky = 0 ;
           "p    " ; p[.,ilo]' ;
           "dif  " ; ((p[.,ihi] - p[.,ilo])./(abs(p[.,ilo])+ _ftol))' ;
           "fp   " ; fp[ilo ihi] ;
           "iter " ; iter ;
           ky = key ;
           if ky ;
              if upper(chrs(ky)) $== "Q" ;
                 break ;
              endif ;
           endif ;
        endif ;
        xp   = (1+alp)*pbar  - alp*p[.,ihi] ;
        fxp  = f(xp) ;
        if fxp < fp[inhi] ;
           if fxp <= fp[ilo] ;
              xxp  = (1+gam)*xp - gam*p[.,ihi] ;
              fxxp = f(xxp) ;
              if fxxp < fp[ilo] ;
              @  "extend" iter ;  @
                 p[.,ihi]   =  xxp ;
                 fp[ihi]    = fxxp ;
                 continue ;
              endif ;
           endif ;
              @  "reflect" iter;  @
              p[.,ihi]  = xp ;
              fp[ihi]   = fxp ;
        else ;
           xp = bet*p[.,ihi] +(1-bet)*pbar ;
           fxp = f(xp) ;
           if fxp < fp[inhi] ;
              @ "shrink 1 dimension" iter ; @
              p[.,ihi] = xp ;
              fp[ihi]  = fxp ;
           else ;
              @ "shrink simplex" iter ; @
              p =  (bet*p + (1-bet)*p[.,ilo]);
           endif ;
        endif ;
    endo ;
    if iter > _maxit ;
       errorlog("nelder: exceeded _maxit") ;
    endif ;
    retp(p[.,ilo], fp[ilo]);
endp ;

/*
**  NN        : neural network
**
**  purpose   : NN is a 3 layer neural network which given input data x,
**              produces an output vector ( matrix?) y.
**
**  syntax    : y = nn(X,A,C) ;
**
**  input     :
**              x :   T x k   matrix of input values
**              A :   k x nH matrix of connection weights between
**                            input layer and hidden layer.
**              C : _nH x 1?  vector of connection weights between input
**                            layer and output layer
**
**  output    : y     : T x 1?  : output of neural netwerk, with
**                                y = G(X*A + b) * C, with G being
**                                the activation function
**
**
**  remarks   :
**               This implementation uses g(x)= 1/(1+exp(-x))
**               as the activation function;
**               change the dimension of C to nH x nO to produce
**               more output values.
**
**
**  date      :  16-11-93
**
**  programmer: G. Draisma
**
*/

proc nn(X,A,c) ;
    retp((1./(1+exp(-(X*A))))*C) ;
endp ;






